'''
Created on 2023-04-06

@author: wf
'''
from tests.basetest import Basetest
from mwdocker.config import MwClusterConfig, Host
import json
from argparse import ArgumentParser
import dataclasses

class TestConfig(Basetest):
    '''
    test the Mediawiki Cluster configuration
    '''
    
    def testDefaults(self):
        """
        test the defaults
        """
        mwClusterConfig=MwClusterConfig()
        expected={'version': '1.39.3', 'smwVersion': None, 
                  'extensionNameList': ['Admin Links', 'Header Tabs', 'SyntaxHighlight', 'Variables'], 
                  'extensionJsonFile': None, 'user': 'Sysop', 'prefix': 'mw', 
                  'password_length': 15, 'password': 'sysop-1234!', 
                  'mySQLRootPassword': None, 'mySQLPassword': None, 
                  'logo': '$wgResourceBasePath/resources/assets/wiki.png',
                   'port': 9080, 'sqlPort': 9306, 'prot': 'http', 
                   'host': Host.get_default_host(), 'script_path': '', 'container_base_name': 'mw-139', 'networkName': 'mwNetwork', 'mariaDBVersion': '10.11', 'forceRebuild': False, 'debug': False, 'verbose': True, 'wikiId': None, 'versions': ['1.35.10','1.38.6', '1.39.3'], 'basePort': 9080}

        mwd=mwClusterConfig.as_dict()
        debug=self.debug
        #debug=True
        if debug:
            print(mwd)
            print(json.dumps(mwd,indent=2))
        self.assertEqual(expected,mwd)
            
    def testArgs(self):
        """
        test command line argument handling
        """
        parser = ArgumentParser()
        mwClusterConfig=MwClusterConfig()
        mwClusterConfig.addArgs(parser)
        argv=["--prot","https"]
        args = parser.parse_args(argv)
        mwClusterConfig.fromArgs(args)
        mwd=dataclasses.asdict(mwClusterConfig)
        debug=self.debug
        #debug=True
        if debug:
            print(json.dumps(mwd,indent=2))
        self.assertEqual("https",mwClusterConfig.prot)
        
    def test_random_password(self):
        """
        test the random password generation
        """
        config=MwClusterConfig()
        for length,chars in [(11,15),(13,18),(15,20)]:
            rp=config.random_password(length)
            debug=self.debug
            if debug:
                print(f"{length} bytes:{len(rp)} chars:{rp}")
            self.assertEqual(chars,len(rp))    


'''
Created on 2023-04-06

@author: wf
'''
import dataclasses
from dataclasses import dataclass, field
import re
import secrets
import socket
from typing import Optional, List
from mwdocker.mw import ExtensionList

class Host:
    """
    Host name getter
    """
    @classmethod    
    def get_default_host(cls) -> str:
        """
        get the default host as the fully qualifying hostname
        of the computer the server runs on
        
        Returns:
            str: the hostname
        """
        host = socket.getfqdn()
        # work around https://github.com/python/cpython/issues/79345
        if host == "1.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.ip6.arpa":
            host = "localhost"  # host="127.0.0.1"
        return host
        
@dataclass
class MwConfig:
    """
    MediaWiki configuration for a Single Wiki
    """
    version:str="1.39.3"
    smwVersion:str=None
    extensionNameList:Optional[List[str]]=field(default_factory=lambda: ["Admin Links","Header Tabs","SyntaxHighlight","Variables"])
    extensionJsonFile:str=None
    user:str="Sysop"
    prefix:str="mw"
    password_length:int = 15
    password:str="sysop-1234!"
    mySQLRootPassword:str=None
    mySQLPassword:str=None
    logo:str="$wgResourceBasePath/resources/assets/wiki.png"
    port:int=9080
    sqlPort:int=9306
    prot:str="http"
    host:str=Host.get_default_host()    
    script_path:str=""
    container_base_name:str=None  
    networkName:str="mwNetwork"
    mariaDBVersion:str="10.11"
    forceRebuild:bool=False
    debug:bool=False
    verbose:bool=True
    wikiId:str=None
    
    def __post_init__(self):
        """
        post initialization configuration
        """
        self.fullVersion=f"MediaWiki {self.version}"
        self.underscoreVersion=self.version.replace(".","_")
        self.shortVersion=self.getShortVersion()
        self.base_url=f"{self.prot}://{self.host}"
        self.url=f"{self.base_url}{self.script_path}:{self.port}"
        if not self.container_base_name:
            self.container_base_name=f"{self.prefix}-{self.shortVersion}"
            
    def as_dict(self)->dict:
        """
        return my fields as a dict
        dataclasses to dict conversion convenienc and information hiding
        
        Returns:
            dict: my fields in dict format
        """
        config_dict=dataclasses.asdict(self)
        return config_dict
        
    def getShortVersion(self,separator=""):
        '''
        get my short version e.g. convert 1.27.7 to 127
        
        Returns:
            str: the short version string
        '''
        versionMatch=re.match("(?P<major>[0-9]+)\.(?P<minor>[0-9]+)",self.version)
        shortVersion=f"{versionMatch.group('major')}{separator}{versionMatch.group('minor')}"
        return shortVersion
    
    def random_password(self,length:int = 15)->str:
        """
        create a random password

        Args:
            length(int): the length of the password

        Returns:
            str:a random password with the given length
        """
        return secrets.token_urlsafe(length)
    
    def getWikiId(self):
        """
        get the wikiId
        
        Returns:
            str: e.g. mw-9080
        """
        if self.wikiId is None:
            wikiId=f"{self.prefix}-{self.port}"
        else:
            wikiId=self.wikiId
        return wikiId
    
    def getExtensionMap(self,extensionNameList:list=None,extensionJsonFile:str=None):
        '''
        get map of extensions to handle
        
        Args:
            extensionNameList(list): a lit of extension names
            extensionJsonFile(str): the name of an extra extensionJsonFile (if any)
        '''
        self.extensionMap={}
        extensionList=ExtensionList.restore()
        if extensionJsonFile is not None:
            extraExtensionList=ExtensionList()
            extraExtensionList.restoreFromJsonFile(extensionJsonFile.replace(".json",""))
            for ext in extraExtensionList.extensions:
                extensionList.extensions.append(ext)
        self.extByName,duplicates=extensionList.getLookup("name")
        if len(duplicates)>0:
            print(f"{len(duplicates)} duplicate extensions: ")
            for duplicate in duplicates:
                print(duplicate.name)
        if extensionNameList is not None:
            self.addExtensions(extensionNameList)
        return self.extensionMap          
    
    def addExtensions(self,extensionNameList):
        """
        add extensions for the given list of extension names
        """
        for extensionName in extensionNameList:
            if extensionName in self.extByName:
                self.extensionMap[extensionName]=self.extByName[extensionName]
            else:
                print(f"warning: extension {extensionName} not known")
        
    def fromArgs(self,args):
        """
        initialize me from the given commmand line arguments
        
        Args:
            args(Namespace): the command line arguments
        """
        self.prefix=args.prefix
        self.container_base_name=args.container_name
        self.extensionNameList=args.extensionNameList
        self.extensionJsonFile=args.extensionJsonFile
        self.forceRebuild=args.forceRebuild
        self.host=args.host
        self.logo=args.logo
        self.mariaDBVersion=args.mariaDBVersion
        # passwords
        self.mySQLRootPassword=args.mysqlPassword
        if not self.mySQLRootPassword:
            self.mySQLRootPassword=self.random_password(self.password_length)
        self.mySQLPassword=self.random_password(self.password_length)    
        self.prot=args.prot
        self.script_path=args.script_path
        self.versions=args.versions
        self.user=args.user
        self.password=args.password
        self.basePort=args.basePort
        self.sqlPort=args.sqlPort
        self.smwVersion=args.smwVersion
        self.verbose=not args.quiet
        self.debug=args.debug
        self.getExtensionMap(self.extensionNameList, self.extensionJsonFile)
   
    def addArgs(self,parser):
        """
        add Arguments to the given parser
        """
        parser.add_argument('-cn','--container_name',default=self.container_base_name,help="set container name (only valid and recommended for single version call)")
        parser.add_argument("-d", "--debug", dest="debug",   action="store_true", help="enable debug mode [default: %(default)s]")
        parser.add_argument('-el', '--extensionList', dest='extensionNameList', nargs="*",default=self.extensionNameList,help="list of extensions to be installed [default: %(default)s]")
        parser.add_argument('-ej', '--extensionJson',dest='extensionJsonFile',default=self.extensionJsonFile,help="additional extension descriptions default: [default: %(default)s]")
        parser.add_argument("-f", "--forceRebuild", action="store_true", help="force rebuilding  [default: %(default)s]")
        parser.add_argument("--host", default=Host.get_default_host(),
                            help="the host to serve / listen from [default: %(default)s]")
        parser.add_argument("--logo", default=self.logo, help="set Logo [default: %(default)s]")
        parser.add_argument('-mv', '--mariaDBVersion', dest='mariaDBVersion',default=self.mariaDBVersion,help="mariaDB Version to be installed [default: %(default)s]")
        parser.add_argument('--mysqlPassword',default=self.mySQLRootPassword, help="set sqlRootPassword [default: %(default)s] - random password if None")
        parser.add_argument('-p','--password',dest='password',default=self.password, help="set password for initial user [default: %(default)s] ")
        parser.add_argument('-pl','--passwordLength',default=self.password_length, help="set the password length for random passwords[default: %(default)s] ")
        parser.add_argument("--prefix",default=self.prefix,help="the container name prefix to use [default: %(default)s]")
        parser.add_argument("--prot",default=self.prot,help="change to https in case [default: %(default)s]")
        parser.add_argument("--script_path",default=self.script_path,help="change to any script_path you might need to set [default: %(default)s]")
        parser.add_argument('-sp', '--sqlBasePort',dest='sqlPort',type=int,default=self.sqlPort,help="set base mySql port 3306 to be exposed - incrementing by one for each version [default: %(default)s]")
        parser.add_argument('-smw','--smwVersion',dest='smwVersion',default=self.smwVersion,help="set SemanticMediaWiki Version to be installed default is None - no installation of SMW")
        parser.add_argument('-u','--user',dest='user',default=self.user, help="set username of initial user with sysop rights [default: %(default)s] ")
        parser.add_argument('-q', '--quiet', help="not verbose [default: %(default)s]",action="store_true")

@dataclass
class MwClusterConfig(MwConfig):
    """
    MediaWiki Cluster configuration for multiple wikis
    """
    versions:Optional[List[str]]=field(default_factory=lambda: ["1.35.10","1.38.6","1.39.3"])
    basePort:int=9080
    
    def addArgs(self,parser):
        """
        add my arguments to the given parser
        """
        super().addArgs(parser) 
        parser.add_argument('-bp', '--basePort',dest='basePort',type=int,default=self.basePort,help="set how base html port 80 to be exposed - incrementing by one for each version [default: %(default)s]")
        parser.add_argument('-vl', '--versionList', dest='versions', nargs="*",default=self.versions,help="mediawiki versions to create docker applications for [default: %(default)s] ")    


'''
Created on 2022-10-25

@author: wf
'''
import re

class MariaDB():
    """
    Maria DB handling
    """
    @classmethod
    def getVersion(cls,versionStr:str)->str:
        """
        get the version from the version String
        
        Args:
            versionStr(str): the version string to check
            
        Return:
            str: the extracted version
        """
        # version is anything which is not a dot at beginning
        # two times may be ending with dash
        version_match=re.search(r"([^.]+[.]+[^.-]+)",versionStr)
        version="?"
        if version_match:
            version=version_match.group(1)
        return version


'''
Created on 2021-08-06
@author: wf
'''
import dataclasses
from mwdocker.docker import DockerApplication
from mwdocker.version import Version
from mwdocker.config import MwClusterConfig
import sys
from argparse import ArgumentParser
from argparse import ArgumentDefaultsHelpFormatter
from mwdocker.logger import Logger
import webbrowser
import traceback

class MediaWikiCluster(object):
    '''
    a cluster of mediawiki docker Applications
    '''
    # https://hub.docker.com/_/mediawiki
    # 2023-01-13
    # MediaWiki Extensions and Skins Security Release Supplement (1.35.9/1.38.4/1.39.1)
    # 2023-02-23 1.39.2 released
    # 2023-04-04 1.39.3 upgrade
    
    def __init__(self,config:MwClusterConfig):
        '''
        Constructor
        
        Args:
            config(MWClusterConfig): the MediaWiki Cluster Configuration to use
        '''
        self.config=config
        self.apps={}       
  
    def createApps(self,home:str=None,withGenerate:bool=True)->dict:
        '''
        create my apps
        
        Args:
            home(str): the home path to use for the docker configuration files
            withGenerate(bool): if True generate the config files
            
        Returns:
            dict(str): a dict of apps by version
        '''  
        for i,version in enumerate(self.config.versions):
            mwApp=self.getDockerApplication(i,version,home)
            if withGenerate:
                mwApp.generateAll(overwrite=self.config.forceRebuild)
            self.apps[version]=mwApp    
        return self.apps
        
    def checkDocker(self)->int: 
        """
        check the Docker environment
        
        print an error message on stderr if check fails
        
        Returns:
            int: exitCode - 0 if ok 1 if failed
        
        """
        errMsg=DockerApplication.checkDockerEnvironment(self.config.debug)
        if errMsg is not None:
            print(errMsg,file=sys.stderr)
            return 1
        return 0
                
    def start(self,forceRebuild:bool=False,withInitDB=True)->int:
        """
        create and start the composer applications
        
        Returns:
            int: exitCode - 0 if ok 1 if failed
        """         
        exitCode=self.checkDocker()  
        if exitCode>0: return exitCode
        
        for version in self.config.versions:
            mwApp=self.apps[version]
            mwApp.start(forceRebuild=forceRebuild,withInitDB=withInitDB)
        return 0
    
    def down(self,forceRebuild:bool=False):
        """
        run docker compose down
        """
        exitCode=self.checkDocker()  
        if exitCode>0: return exitCode
        for _i,version in enumerate(self.config.versions):
            mwApp=self.apps[version]
            mwApp.down(forceRebuild)
    
    def listWikis(self)->int:
        """
        list my wikis
        
        Returns:
            int: exitCode - 0 if ok 1 if failed
        """
        exitCode=self.checkDocker()  
        if exitCode>0: return exitCode
        for i,version in enumerate(self.config.versions):
            mwApp=self.apps[version]
            mw,db=mwApp.getContainers()
            if mw and db:
                l_str="found"
            else:
                l_str="missing"
            msg=f"{i+1}:{version} {l_str}"
            print(msg)
        return exitCode
        
        
    def check(self)->int:
        """
        check the composer applications
        
        Returns:
            int: exitCode - 0 if ok 1 if failed
        """         
        exitCode=self.checkDocker()  
        if exitCode>0: return exitCode
        
        for i,version in enumerate(self.config.versions):
            mwApp=self.apps[version]
            msg=f"{i+1}:checking {version} ..."
            print(msg)
            exitCode=mwApp.check()
        return exitCode
            
    def close(self):
        """
        close my apps
        """
        for mwApp in self.apps.values():
            mwApp.close()
            
    def getDockerApplication(self,i:int,version:str,home:str=None):
        '''
        get the docker application for the given version index and version
        
        Args:
            i(int): the index of the version
            version(str): the mediawiki version to use
        
        Returns:
            DockerApplication: the docker application
        '''
        # please note that we are using the subclass MwClusterConfig here although
        # we only need the superclass MwConfig - we let inheritance work here for us but
        # have to ignore the superfluous fields
        appConfig=dataclasses.replace(self.config)
        appConfig.extensionMap=self.config.extensionMap.copy()
        appConfig.version=version
        appConfig.port=self.config.basePort+i
        appConfig.sqlPort=self.config.sqlPort+i   
        # let post_init create a new container_base_name
        appConfig.container_base_name=None
        appConfig.__post_init__()            
        mwApp=DockerApplication(config=appConfig,home=home)
        return mwApp

DEBUG=False

def main(argv=None): # IGNORE:C0111
    '''main program.'''

    if argv is None:
        argv = sys.argv[1:]

    program_name = "mwcluster"
    program_version = f"v{Version.version}"
    program_build_date = str(Version.updated)
    program_version_message = f'{program_name} ({program_version},{program_build_date})'
    program_license = Version.license
    
    try:
        # Setup argument parser
        parser = ArgumentParser(description=program_license, formatter_class=ArgumentDefaultsHelpFormatter)
        mwClusterConfig=MwClusterConfig()
        mwClusterConfig.addArgs(parser)
        parser.add_argument("--about", help="show about info [default: %(default)s]", action="store_true")
        parser.add_argument("--create", action="store_true", help="create wikis [default: %(default)s]")
        parser.add_argument("--down", action="store_true", help="shutdown wikis [default: %(default)s]")
        parser.add_argument("--check",action="store_true",help="check the wikis [default: %(default)s]")
        parser.add_argument("--list",action="store_true",help="list the wikis [default: %(default)s]")
        parser.add_argument('-V', '--version', action='version', version=program_version_message)
        args = parser.parse_args(argv)
        if args.about:
            print(program_version_message)
            print(f"see {Version.doc_url}")
            webbrowser.open(Version.doc_url)
        else:
            action=None
            withGenerate=False
            if args.check: 
                action="checking docker access" 
            elif args.create: 
                action="creating docker compose applications" 
                withGenerate=True
            elif args.list:
                action="listing docker compose wiki applications"
            elif args.down:
                action="running docker compose down"
            if not action:
                parser.print_usage()
            else:
                print(f"{action} for mediawiki versions {args.versions}")
                # create a MediaWiki Cluster
                mwClusterConfig.fromArgs(args)
                mwCluster=MediaWikiCluster(config=mwClusterConfig)
                mwCluster.createApps(withGenerate=withGenerate)
                if args.check:
                    return mwCluster.check()
                elif args.create:
                    return mwCluster.start(forceRebuild=args.forceRebuild)
                elif args.list:
                    return mwCluster.listWikis()
                elif args.down:
                    return mwCluster.down(forceRebuild=args.forceRebuild)
    except KeyboardInterrupt:
        ### handle keyboard interrupt ###
        return 1
    except Exception as e:
        if DEBUG:
            raise(e)
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        if args is None:
            print("args could not be parsed")
        elif args.debug:
            print(traceback.format_exc())
        return 2
            
if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-d")
    sys.exit(main())
    


'''
Created on 2021-08-06

@author: wf
'''
from python_on_whales import docker
import os
import platform
import datetime
import pprint
import time
import traceback
import typing
from jinja2 import Environment, FileSystemLoader
from jinja2.exceptions import TemplateNotFound
import mysql.connector
from mysql.connector import Error
from pathlib import Path
from wikibot3rd.wikiuser import WikiUser
from mwdocker.logger import Logger
from mwdocker.config import MwClusterConfig
from mwdocker.html_table import HtmlTables
from mwdocker.mariadb import MariaDB
from lodstorage.lod import LOD
from dataclasses import dataclass

class DockerMap():
    '''
    helper class to convert lists of docker elements to maps for improved
    lookup functionality
    '''
    @staticmethod
    def getContainerMap():
        '''
        get a map/dict of containers by container name
        '''
        containerMap={}
        for container in docker.container.list():
            containerMap[container.name]=container
            pass
        return containerMap
    
class DockerContainer():
    """
    helper class for docker container info
    """
    
    def __init__(self,name,kind,container):
        """
        constructor
        """
        self.name=name
        self.kind=kind
        self.container=container
       
    def check(self):
        """
        check the given docker container
        
        print check message and Return if container is running
        
        Args:
            dc: the docker container
        
        Returns:
            bool: True if the container is not None
        """
        ok=self.container.state.running
        msg=f"mediawiki {self.kind} container {self.name}"
        return Logger.check_and_log(msg, ok) 
    
    def getHostPort(self,local_port:int=80)->int:
        """
        get the host port for the given local port
        
        Args:
            local_port(int): the local port to get the mapping for
            
        Returns:
            int: the  host port or None
        """
        host_port=None
        pb_dict=self.container.host_config.port_bindings
        p_local=f"{local_port}/tcp"
        if p_local in pb_dict:
            pb=pb_dict[p_local][0]
            host_port=pb.host_port
        return host_port
    
@dataclass
class DBStatus():
    """
    the Database Status
    """
    attempts: int
    max_tries: int
    ok: bool
    msg:str
    ex:typing.Optional[Exception]=None
        
class DockerApplication(object):
    '''
    MediaWiki Docker image
    '''

    def __init__(self,
                 config:MwClusterConfig,
                 home:str=None):
        '''
        Constructor
        
        Args:
            config: MwClusterConfig,
            home: the home directory to use
        '''
        self.config=config
        # branch as need for git clone e.g. https://gerrit.wikimedia.org/g/mediawiki/extensions/MagicNoCache
        self.branch=f"REL{self.config.getShortVersion('_')}"
        self.composerVersion=1
        if self.config.shortVersion>="139":
            self.composerVersion=2
        # jinja and docker prerequisites
        self.env=self.getJinjaEnv()
        if home is None:
            home = str(Path.home())
        self.dockerPath=f'{home}/.pymediawikidocker/{self.config.container_base_name}' 
        os.makedirs(self.dockerPath,exist_ok=True)
        
        self.getContainers()
        self.dbConn=None
        self.database="wiki"
        self.dbUser="wikiuser"
        self.wikiUser=None
       
    @staticmethod 
    def checkDockerEnvironment(debug:bool=False)->str:
        """
        check the docker environment
        
        Args:
            debug(bool): if True show debug information
            
        Returns:
            str: an error message or None
        """
        errMsg=None
        if not docker.compose.is_installed():
            errMsg="""docker composer up needs to be working"""
        os_path=os.environ["PATH"]
        paths=["/usr/local/bin"]
        for path in paths:
            if os.path.islink(f"{path}/docker"):
                if not path in os_path:
                    os.environ["PATH"]=f"{os_path}{os.pathsep}{path}"
                    if debug:
                        print(f"""modified PATH from {os_path} to \n{os.environ["PATH"]}""")
        return errMsg
    
    def check(self)->int:
        """
        check me
        
        Returns:
            int: exitCode: 0 if ok, 1 if not ok
        """
        DockerApplication.checkDockerEnvironment(self.config.debug) 
        exitCode=0
        mw,db=self.getContainers()
        if not mw:
            print("mediawiki container missing")
            exitCode=1
        if  not db:
            print("database container missing")
            exitCode=1
        if mw and db and mw.check() and db.check():
            host_port=mw.getHostPort(80)
            if host_port:
                Logger.check_and_log_equal(f"port binding",host_port,"expected  port",str(self.config.port))
                url=self.config.url
                # fix url to local port
                # @TODO isn't this superfluous / has no effect ...?
                url=url.replace(str(self.config.port),host_port)
                version_url=f"{url}/index.php/Special:Version"
                
                ok=self.checkWiki(version_url)
                if not ok:
                    exitCode=1
            else:
                self.log(f"port binding for port 80 missing",False)
                exitCode=1
            pass
        return exitCode
    
    def checkWiki(self,version_url:str)->bool:
        """
        check this wiki against the content of the given version_url
        """
        print(f"Checking {version_url} ...")
        ok=True
        try:
            html_tables=HtmlTables(version_url)
            tables=html_tables.get_tables("h2")
            if self.config.debug:
                p = pprint.PrettyPrinter(indent=2)
                p.pprint(tables)
            ok=ok and Logger.check_and_log("Special Version accessible ...", "Installed software" in tables)
            if ok:
                software=tables["Installed software"]
                software_map,_dup=LOD.getLookup(software, "Product", withDuplicates=False)
                mw_version=software_map["MediaWiki"]["Version"]
                ok=ok and Logger.check_and_log_equal("Mediawiki Version",mw_version,"expected ",self.config.version)
                db_version_str=software_map["MariaDB"]["Version"]
                db_version=MariaDB.getVersion(db_version_str)
                ok=ok and Logger.check_and_log(f"Maria DB Version {db_version} fitting expected {self.config.mariaDBVersion}?",self.config.mariaDBVersion.startswith(db_version))
                pass
        except Exception as ex:
            ok=Logger.check_and_log(str(ex), False)
        return ok
    
    def getContainerName(self,kind:str,separator:str):
        """
        get my container Name
        """
        containerName=f"{self.config.container_base_name}{separator}{kind}"
        return containerName
    
    def getContainers(self):
        """
        get my containers
        
        Returns:
            Tuple(
        """
        self.dbContainer=None
        self.mwContainer=None
        containerMap=DockerMap.getContainerMap()
        for separator in ["-","_"]:
            dbContainerName=self.getContainerName("db", separator)
            mwContainerName=self.getContainerName("mw", separator)
            if dbContainerName in containerMap:
                self.dbContainer=DockerContainer(dbContainerName,"database",containerMap[dbContainerName])
            if mwContainerName in containerMap:               
                self.mwContainer=DockerContainer(mwContainerName,"webserver",containerMap[mwContainerName])   
        return self.mwContainer,self.dbContainer   
            
    def getJinjaEnv(self):
        '''
        get a Jinja2 environment
        '''
        scriptdir=os.path.dirname(os.path.realpath(__file__))
        resourcePath=os.path.realpath(f"{scriptdir}/resources")
        template_dir = os.path.realpath(f'{resourcePath}/templates')
        #print(f"jinja template directory is {template_dir}")
        env = Environment(loader=FileSystemLoader(template_dir))
        return env
    
    def initDB(self):
        '''
        initialize my SQL database
        '''
        # restore the mySQL dump data
        self.execute("/tmp/initdb.sh")
        # update the database e.g. to initialize Semantic MediaWiki tables
        self.execute("/tmp/update.sh")
        # add an initial sysop user as specified
        self.execute("/tmp/addSysopUser.sh")
            
    def installExtensions(self):
        '''
        install all extensions
        '''
        self.execute("/tmp/installExtensions.sh")
        
    def startUp(self):
        '''
        run startUp scripts
        '''
        self.execute("/root/addCronTabEntry.sh")
            
    def createWikiUser(self,wikiId:str=None,store:bool=False):
        '''
        create my wikiUser and optionally save it
        
        Args:
           store(bool): if True save my user data to the relevant ini File
        '''
        if not wikiId:
            wikiId=f"{self.config.container_base_name}"
        userDict={
            "wikiId":f"{wikiId}",
            "url": f"{self.config.base_url}",
            "scriptPath": f"{self.config.script_path}",
            "user": f"{self.config.user}",
            "email":"noreply@nouser.com",
            "version": f"{self.config.fullVersion}",
            "password": f"{self.config.password}"
        }
        wikiUser=WikiUser.ofDict(userDict,encrypted=False)
        if store:
            wikiUser.save()
        return wikiUser
    
    def execute(self,command_str:str):
        '''
        execute the given command string
        
        Args:
            command_str: str - a command string to be executed ...
        '''
        if self.mwContainer:
            if self.config.verbose:
                print(f"Executing docker command {command_str}")
            docker.execute(container=self.mwContainer.container,command=[command_str])
        else:
            mwContainerNameDash=self.getContainerName("mw", "-")
            mwContainerNameUnderscore=self.getContainerName("mw", "_")
            errMsg=f"no mediawiki Container {mwContainerNameDash} or {mwContainerNameUnderscore} for {self.name} activated by docker compose\n- you might want to check the separator character used for container names for your platform {platform.system()}"
            raise Exception(f"{errMsg}")
    
    def close(self):
        """
        close the database
        """
        self.dbClose()
    
    def sqlQuery(self,query):
        """
        run the given SQL query
        """
        if self.dbConn and self.dbConn.is_connected():
            cursor = self.dbConn.cursor()
            cursor.execute(query)
            rows = cursor.fetchall()
            cursor.close()
            return rows
        else:
            if self.config.verbose:
                print (f"Connection to {self.database} on {self.config.host} with user {self.dbUser} not established" )
            return None
        
    def dbClose(self):
        """
        close the database connection
        """
        if self.dbConn and self.dbConn.is_connected():
            self.dbConn.close()
        
    def dbConnect(self,timeout:int=10):
        """
        connect to the database and return the connection
        
        Args:
            timeout(int): number of seconds for timeout
            
        Returns:
            the connection
        """
        if self.dbConn is None:
            try:
                self.dbConn = mysql.connector.connect(host=self.config.host,
                                 database=self.database,
                                 user=self.dbUser,
                                 port=self.config.sqlPort,
                                 password=self.config.mySQLPassword,
                                 connection_timeout=timeout)
        
            except Error as e :
                errMsg=str(e)
                print (f"Connection to {self.database} on {self.config.host} with user {self.dbUser} failed error: {errMsg}" )
                if "Access denied" in errMsg:
                    raise e
        return self.dbConn
    
    def doCheckDBConnection(self,dbStatus:DBStatus,timeout:int=10):
        """
        check the database connection of this application
        
        Args:
            timeout(int): how many seconds to wait
            
        Returns:
            DBStatus
        """          
        dbStatus.attempts+=1
        self.dbConnect(timeout=timeout)
        if self.dbConn and self.dbConn.is_connected():
            rows=self.sqlQuery("select database();")
            dbStatus.ok=True
            if self.config.verbose:
                print (f"{dbStatus.msg} established database returns: {rows[0]}")     
    
    def checkDBConnection(self,timeout:float=10,initialSleep:float=2.5,maxTries:int=7)->DBStatus:
        """
        check database connection with retries
        
        Args:
            timeout(float): number of seconds for timeout
            initialSleep(float): number of seconds to initially wait/sleep
            maxTries(int): maximum number of retries before giving up between each try a sleep is done that starts
            with 0.5 secs and doubles on every retry
            
        Returns:
            dbStatus: the status
        """ 
        conn_msg=f"SQL-Connection to {self.database} on {self.config.host} port {self.config.sqlPort} with user {self.dbUser}"
        dbStatus=DBStatus(attempts=0,ok=False,msg=conn_msg,max_tries=maxTries)
        if self.config.debug:
            print (f"Trying {dbStatus.msg} with max {maxTries} tries and {timeout}s timeout per try - initial sleep {initialSleep}s")
        time.sleep(initialSleep)
        sleep=0.5
        while not dbStatus.ok and dbStatus.attempts<=maxTries:
            try:
                self.doCheckDBConnection(dbStatus,timeout=timeout)
                if not dbStatus.ok:
                    if self.config.verbose:
                        print(f"Connection attempt #{dbStatus.attempts}/{dbStatus.max_tries} failed will retry in {sleep} secs" )   
                    # wait before trying
                    time.sleep(sleep)
                    sleep=sleep*2
            except Exception as ex:
                dbStatus.ex=ex
                if self.config.verbose:
                    print(f"Connection attempt #{dbStatus.attempts} failed with exception {str(ex)} - will not retry ...")
                if self.config.debug:
                    print(traceback.format_exc())
                break
        return dbStatus
    
    def optionalWrite(self,targetPath:str,content:str,overwrite:bool=False):
        """
        optionally Write the modified content to the given targetPath
        
        Args:
            targetPath(str): the path to write the content to
            content(str): the content to write
            overwrite(bool): if True overwrite the existing content
        """
        if not overwrite and os.path.isfile(targetPath):
            if self.config.verbose:
                print(f"{targetPath} already exists!")
            return
        with open(targetPath, "w") as targetFile:
                targetFile.write(content)
    
    def generate(self,templateName:str,targetPath:str,overwrite:bool=False,**kwArgs):
        '''
        generate file at targetPath using the given templateName
        
        Args:
            templateName(str): the Jinja2 template to use
            targetPath(str): the path to the target file
            overwrite(bool): if True overwrite existing files
            kwArgs(): generic keyword arguments to pass on to template rendering
        '''
        try:
            template = self.env.get_template(templateName)
            timestamp=datetime.datetime.now().isoformat()
            content=template.render(mwVersion=self.config.version,mariaDBVersion=self.config.mariaDBVersion,port=self.config.port,sqlPort=self.config.sqlPort,smwVersion=self.config.smwVersion,timestamp=timestamp,**kwArgs)
            self.optionalWrite(targetPath, content, overwrite)

        except TemplateNotFound:
            print(f"no template {templateName} for {self.config.name} {self.config.version}")     
    
    def getComposerRequire(self):
        '''
        get the json string for the composer require e.g. composer.local.json
        '''
        requires=[]
        for ext in self.config.extensionMap.values():
            if hasattr(ext,"composer"):
                # get the composer statement
                composer=ext.composer
                requires.append(composer)
        indent="     "
        delim="" if len(requires)==0 else ",\n"
        requireList=""
        if self.config.smwVersion:
            requireList+=f'{indent}"mediawiki/semantic-media-wiki": "~{self.config.smwVersion}"{delim}'
        for i,require in enumerate(requires):
            delim="" if i>=len(requires)-1 else ",\n"
            requireList+=f"{indent}{require}{delim}"
        requireJson=f"""{{   
  "require": {{
{requireList} 
  }} 
}}"""
        return requireJson
        
    def genComposerRequire(self,composerFilePath,overwrite:bool=False):
        '''
        gen the composer.local.json require file
        
        Args:
            composerFilePath(str): the name of the file to generate
        '''
        requireJson=self.getComposerRequire()
        self.optionalWrite(composerFilePath, requireJson, overwrite)
             
    def generateAll(self,overwrite:bool=False):
        '''
        generate all files needed for the docker handling
        
        Args:
            overwrite(bool): if True overwrite the existing files
        '''
        self.generate("mwDockerfile",f"{self.dockerPath}/Dockerfile",composerVersion=self.composerVersion,overwrite=overwrite)
        self.generate("mwCompose.yml",f"{self.dockerPath}/docker-compose.yml",mySQLRootPassword=self.config.mySQLRootPassword,mySQLPassword=self.config.mySQLPassword,container_base_name=self.config.container_base_name,overwrite=overwrite)
        self.generate(f"mwLocalSettings{self.config.shortVersion}.php",f"{self.dockerPath}/LocalSettings.php",mySQLPassword=self.config.mySQLPassword,hostname=self.config.host,extensions=self.config.extensionMap.values(),mwShortVersion=self.config.shortVersion,logo=self.config.logo,overwrite=overwrite)
        self.generate(f"mwWiki{self.config.shortVersion}.sql",f"{self.dockerPath}/wiki.sql",overwrite=overwrite)
        self.generate(f"addSysopUser.sh",f"{self.dockerPath}/addSysopUser.sh",user=self.config.user,password=self.config.password,overwrite=overwrite)
        self.generate(f"installExtensions.sh",f"{self.dockerPath}/installExtensions.sh",extensions=self.config.extensionMap.values(),branch=self.branch,overwrite=overwrite)
        self.genComposerRequire(f"{self.dockerPath}/composer.local.json",overwrite=overwrite)
        for fileName in ["addCronTabEntry.sh","startRunJobs.sh","initdb.sh","update.sh","phpinfo.php","upload.ini","plantuml.sh"]:
            self.generate(f"{fileName}",f"{self.dockerPath}/{fileName}",overwrite=overwrite)
        
    def down(self,forceRebuild:bool=False):
        """
        run docker compose down
        
        see https://docs.docker.com/engine/reference/commandline/compose_down/
        and https://gabrieldemarmiesse.github.io/python-on-whales/sub-commands/compose/#down
        
        """
        DockerApplication.checkDockerEnvironment(self.config.debug) 
        # change directory so that docker CLI will find the relevant dockerfile and docker-compose.yml
        if self.config.verbose:
            print(f"running docker compose down for {self.config.container_base_name} {self.config.version} docker application ...")
        # remember current directory 
        cwd = os.getcwd()    
        os.chdir(self.dockerPath)
        docker.compose.down(volumes=forceRebuild) 
        # switch back to previous current directory
        os.chdir(cwd)
        
    def up(self,forceRebuild:bool=False):
        """
        start this docker application
        
        Args: 
            forceRebuild(bool): if true stop and remove the existing containers
        """     
        DockerApplication.checkDockerEnvironment(self.config.debug)       
        if self.config.verbose:
            print(f"starting {self.config.container_base_name} {self.config.version} docker application ...")
        if forceRebuild:
            for docker_container in [self.dbContainer,self.mwContainer]:
                if docker_container is not None:
                    container=docker_container.container
                    try:
                        container_name=container.name
                        if self.config.verbose:
                            print(f"stopping and removing container {container_name}")
                    except Exception as container_ex:
                        container=None
                    if container:
                        try:
                            container.stop()
                        except Exception as stop_ex:
                            if self.config.verbose:
                                print(f"stop failed with {str(stop_ex)}")
                            pass
                        try:
                            container.remove()
                        except Exception as remove_ex:
                            if self.config.verbose:
                                print(f"removed failed with {str(remove_ex)}")
                            pass
                    pass

        # remember current directory 
        cwd = os.getcwd()    
     
        # change directory so that docker CLI will find the relevant dockerfile and docker-compose.yml
        os.chdir(self.dockerPath)
        #project_config = docker.compose.config()
        if forceRebuild:
            docker.compose.build()
        # run docker compose up
        # this might take a while e.g. downloading
        # run docker compose up
        docker.compose.up(detach=True,force_recreate=forceRebuild)      
        # switch back to previous current directory
        os.chdir(cwd)
    
        return self.getContainers()
        
    def start(self,forceRebuild:bool=False,withInitDB=True):
        """
        start my containers
        
        Args:
            forceRebuild(bool): if True force rebuilding
            withInitDB(bool): if True intialize my database
        """         
        self.up(forceRebuild=forceRebuild) 
        if withInitDB:
            if self.config.verbose:
                print("Initializing MediaWiki SQL tables")
            dbStatus=self.checkDBConnection()
            if dbStatus.ok:
                # first install extensions
                self.installExtensions()
                # then create and fill database and update it
                self.initDB()
                # then run startUp scripts
                self.startUp()